#cython: boundscheck=False
# -*- coding: utf-8 -*-
"""
cutil.pyx

Created by Stephan Hügel on 2016-07-31

This file is part of pypolyline.

The MIT License (MIT)

Copyright (c) 2016 Stephan Hügel

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""
__author__ = u"Stephan Hügel"

import numpy as np
from pypolyline_p cimport (
    ExternalArray,
    InternalArray,
    decode_polyline_ffi,
    encode_coordinates_ffi,
    drop_float_array,
    drop_cstring
    )

def encode_coordinates(coords, int precision):
    """
    Encode coordinates as a Polyline.
    Input: a list of lat, lon coordinates, and a precision int (5 for Google, 6 for OSM-derived).
    Output: an encoded Polyline string.

    Example: encode_coordinates([[38.5, -120.2], [40.7, -120.95], [43.252, -126.453]], 5)
    Result: "_p~iF~ps|U_ulLnnqC_mqNvxq`@"

    """
    cdef double[:,::1] ncoords = np.array(coords, dtype=np.float64)
    cdef ExternalArray coords_ffi
    coords_ffi.data = <void*>&ncoords[0, 0]
    coords_ffi.len = ncoords.shape[0]
    cdef char* result = encode_coordinates_ffi(coords_ffi, precision)
    cdef bytes polyline = result
    drop_cstring(result)
    return polyline

def decode_polyline(bytes polyline, int precision):
    """
    Decode an encoded Polyline to coordinates.
    Input: a Polyline string, and a precision int (5 for Google, 6 for OSM-derived).
    Output: a list of lat, lon coordinates.

    Example: decode_polyline(_p~iF~ps|U_ulLnnqC_mqNvxq`@, 5)
    Result: [[38.5, -120.2], [40.7, -120.95], [43.252, -126.453]]

    """
    cdef char* to_send = polyline
    cdef InternalArray result = decode_polyline_ffi(to_send, precision)
    cdef double* incoming_ptr = <double*>(result.data)
    cdef double[:, ::1] view = <double[:result.len,:2:1]>incoming_ptr
    cdef coords = np.copy(view).tolist()
    drop_float_array(result)
    return coords



cdef extern from "header.h":
    struct InternalArray:
        void* data
        size_t len

    struct ExternalArray:
        void* data
        size_t len

    cdef InternalArray decode_polyline_ffi(char* polyline, int precision);
    cdef char* encode_coordinates_ffi(ExternalArray, int precision);
    cdef void drop_float_array(InternalArray coords);
    cdef void drop_cstring(char* polyline)



import numpy as np

import cgpolyencode
encoder = cgpolyencode.GPolyEncoder()

# London bounding box
N = 51.691874116909894
E = 0.3340155643740321
S = 51.28676016315085
W = -0.5103750689005356

num_coords = 1000
coords = zip(
    np.random.uniform(S, N, [num_coords]),
    np.random.uniform(W, E, [num_coords])
    ) 

if __name__ == "__main__":
    for x in range(500):
        encoder.encode(coords)



"""
util.py

Created by Stephan Hügel on 2016-07-19

This file is part of pypolyline.

The MIT License (MIT)

Copyright (c) 2016 Stephan Hügel

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import os
from sys import platform, version_info
from ctypes import (
    Structure,
    POINTER,
    c_void_p,
    c_size_t,
    c_double,
    c_uint32,
    c_char_p,
    cast,
    cdll,
)
import numpy as np

# Python 3 check
from subprocess import getoutput as spop

__author__ = "Stephan Hügel"

file_path = os.path.dirname(__file__)

prefix = {"win32": ""}.get(platform, "lib")
extension = {"darwin": ".dylib", "win32": ".dll"}.get(platform, ".so")
fpath = {"darwin": "", "win32": ""}.get(platform, os.path.join(file_path, ".libs"))


py3 = True

try:
    lib = cdll.LoadLibrary(os.path.join(file_path, prefix + "polylineffi" + extension))
except OSError:
    # the Rust lib's been grafted by manylinux1
    if not py3:
        fname = spop(["ls", fpath]).split()[0]
    else:
        fname = spop(["ls %s" % fpath]).split()[0]
    lib = cdll.LoadLibrary(os.path.join(file_path, ".libs", fname))


class EncodingError(Exception):
    def __init__(self, message):
        super().__init__(message)


class DecodingError(Exception):
    def __init__(self, message):
        super().__init__(message)


class _FFIArray(Structure):
    """
    Convert sequence of float lists to a C-compatible void array
    example: [[1.0, 2.0], [3.0, 4.0]]

    """

    _fields_ = [("data", c_void_p), ("len", c_size_t)]

    @classmethod
    def from_param(cls, seq):
        """Allow implicit conversions"""
        return seq if isinstance(seq, cls) else cls(seq)

    def __init__(self, seq, data_type=c_double):
        array = np.array(seq, dtype=np.float64)
        self._buffer = array.data
        self.data = cast(array.ctypes.data_as(POINTER(data_type)), c_void_p)
        self.len = len(seq)


class _CoordResult(Structure):
    """Container for returned FFI coordinate data"""

    _fields_ = [("coords", _FFIArray)]


class _PolylineResult(Structure):
    """Container for returned FFI Polyline data"""

    _fields_ = [("line", c_void_p)]


def _void_array_to_nested_list(res, _func, _args):
    """Dereference the FFI result to a list of coordinates"""
    try:
        shape = res.coords.len, 2
        ptr = cast(res.coords.data, POINTER(c_double))
        array = np.ctypeslib.as_array(ptr, shape)
        if np.isnan(np.sum(array)):
            raise DecodingError("Your Polyline was not valid and could not be decoded")
        return array.tolist()
    finally:
        _drop_array(res.coords)


def _void_array_to_string(res, _func, _args):
    """Dereference the FFI result to a utf8 polyline"""
    try:
        result = cast(res.line, c_char_p)
        polyline = bytes(result.value)
        if polyline.startswith(b"Latitude"):
            raise EncodingError(
                "%s. Latitudes must be between -90.0 and 90.0" % polyline
            )
        elif polyline.startswith(b"Longitude"):
            raise EncodingError(
                "%s. Longitudes must be between -180.0 and 180.0" % polyline
            )
        return polyline
    finally:
        _drop_cstring(res.line)


decode_polyline = lib.decode_polyline_ffi
decode_polyline.argtypes = (c_char_p, c_uint32)
decode_polyline.restype = _CoordResult
decode_polyline.errcheck = _void_array_to_nested_list
decode_polyline.__doc__ = """
    Decode an encoded Polyline to coordinates.
    Input: a Polyline string, and a precision int (5 for Google, 6 for OSM-derived).
    Output: a list of lon, lat coordinates.

    Example: decode_polyline(_p~iF~ps|U_ulLnnqC_mqNvxq`@, 5)
    Result: [[-120.2, 38.5], [-120.95, 40.7], [-126.453, 43.252]]

    Incorrect Polyline input will throw util.DecodingError

 """

encode_coordinates = lib.encode_coordinates_ffi
encode_coordinates.argtypes = (_FFIArray, c_uint32)
encode_coordinates.restype = _PolylineResult
encode_coordinates.errcheck = _void_array_to_string
encode_coordinates.__doc__ = """
    Encode coordinates as a Polyline.
    Input: a list of lon, lat coordinates, and a precision int (5 for Google, 6 for OSM-derived).
    Output: an encoded Polyline string.

    Example: encode_coordinates([[-120.2, 38.5], [-120.95, 40.7], [-126.453, 43.252]], 5)
    Result: "_p~iF~ps|U_ulLnnqC_mqNvxq`@"

    Incorrect coordinate input will throw util.EncodingError

"""

# Free FFI-allocated memory
_drop_cstring = lib.drop_cstring
_drop_cstring.argtypes = (c_void_p,)
_drop_cstring.restype = None

_drop_array = lib.drop_float_array
_drop_array.argtypes = (_FFIArray,)
_drop_array.restype = None



#!/usr/bin/env python
"""
setup.py

Created by Stephan Hügel on 2016-07-25
"""

import sys
from setuptools import setup, Extension
import numpy
from Cython.Build import cythonize


# # Set dynamic RPATH differently, depending on platform
ldirs = []
ddirs = []
if "linux" in sys.platform:
    # from http://stackoverflow.com/a/10252190/416626
    # the $ORIGIN trick is not perfect, though
    ldirs = ["-Wl,-rpath", "-Wl,$ORIGIN/"]
    platform_lib = "libpolylineffi.so"
if sys.platform == "darwin":
    # You must compile your binary with rpath support for this to work
    # RUSTFLAGS="-C rpath" cargo build --release
    platform_lib = "libpolylineffi.dylib"
    ldirs = ["-Wl,-rpath", "-Wl,@loader_path/"]
if sys.platform == "win32":
    ddirs = ["src/pypolyline/header.h"]
    platform_lib = "polylineffi.dll"


extension = Extension(
    name="pypolyline.cutil",
    sources=["src/pypolyline/cutil.pyx"],
    libraries=["polylineffi"],
    depends=ddirs,
    language="c",
    include_dirs=["src/pypolyline", numpy.get_include()],
    library_dirs=["src/pypolyline"],
    extra_link_args=ldirs,
)

extensions = cythonize(
    [
        extension,
    ],
    compiler_directives={"language_level": "3"},
)

setup(
    package_data={
        "pypolyline": [platform_lib],
    },
    ext_modules=[extension],
)



#!/usr/bin/env python
"""
Standalone benchmark runner
"""

import cProfile
import pstats
import profile
import numpy as np

print("Running Rust, Python, and C++ benchmarks. 1000 points, 500 runs.\n")

# calibrate
print("Calibrating system")
pr = profile.Profile()
calibration = np.mean([pr.calibrate(100000) for x in range(5)])
# add the bias
profile.Profile.bias = calibration
print("Calibration complete, running benchmarks\n")
bmarks = [
    ('benches/benchmark_rust.py', 'benches/output_stats_rust', 'Rust + Cython'),
    ('benches/benchmark_python.py', 'benches/output_stats_python', 'Pure Python'),
]

results = []
for benchmark in bmarks:
    with open(benchmark[0], 'rb') as f:
        cProfile.run(f.read(), benchmark[1])
        results.append(pstats.Stats(benchmark[1]))

for i, benchmark in enumerate(bmarks):
    print("%s Benchmark\n" % benchmark[2])
    results[i].sort_stats('cumulative').print_stats(3)



import numpy as np
from pypolyline.cutil import encode_coordinates

# London bounding box
N = 51.691874116909894
E = 0.3340155643740321
S = 51.28676016315085
W = -0.5103750689005356

num_coords = 1000
coords = list(zip(
    np.random.uniform(S, N, [num_coords]),
    np.random.uniform(W, E, [num_coords])
    ))

if __name__ == "__main__":
    for x in range(500):
        encode_coordinates(coords, 5)
